;;===========================================================================
;; GingerBreadMan.asm - minimal D3D application drawing gingerbread fractal
;;

	INCLUDE GingerBreadMan.Inc

;; width and height of frame buffer
FRAME_WIDTH		EQU 320
FRAME_HEIGHT	EQU 240


_DATA SEGMENT
	wc WNDCLASSEX <SIZEOF WNDCLASSEX,CS_CLASSDC,OFFSET WndProc,0,0,0,0,0,0,0,OFFSET APP_CLASS,0>
	hInst EQU <wc.hInstance>

	msg MSG <>

	;; device window and interface
	g_window	EQU <g_presentation.hDeviceWindow>
	g_d3d		LPDIRECT3D8 0
	g_device	LPDIRECT3DDEVICE8 0

	;; device presentation parameters
	g_presentation D3DPRESENT_PARAMETERS <FRAME_WIDTH, FRAME_HEIGHT, D3DFMT_X8R8G8B8, 1,\
		D3DMULTISAMPLE_NONE, D3DSWAPEFFECT_COPY, NULL, TRUE,\
		TRUE, D3DFMT_D16,0,0,0>
		;; width, height, format, number of back buffers
		;; multisample type, swap effect, device window, windowed
		;; auto depth/stencil surface, depth/stencil surface format
_DATA ENDS


CONST SEGMENT
	APP_CLASS EQU <APP_NAME>
	APP_NAME db "Ginger Bread Man Fractal",0
		ALIGN
CONST ENDS


_TEXT SEGMENT


vertex_s STRUCT
	m_x			REAL4 ?
	m_y			REAL4 ?
	m_z			REAL4 ?
	m_recip_w	REAL4 ?
	m_diffuse	DWORD ?
vertex_s ENDS

	Frame_Draw PROC uses ebx esi edi
		or g_device,0
		je @Bye

		NUM_POINTS EQU 20000

		_BSS SEGMENT
			points vertex_s NUM_POINTS dup (<>)
		_BSS ENDS

		_DATA SEGMENT
			x0 REAL8 -0.1
			y0 REAL8 0.0
			__FRAME_WIDTH REAL4 @CatStr(%FRAME_WIDTH,<.0>)
			__FRAME_HEIGHT REAL4 @CatStr(%FRAME_HEIGHT,<.0>)
			___1f REAL4 1.0
			__255 REAL4 255.0
			__65536 REAL4 65536.0
		_DATA ENDS

		fld y0		; min y = st5
		fld x0		; min x = st4
		fld st(1)	; max y = st3
		fld st(1)	; max x = st2
		fld st(1)	; y = st1
		fld st(1)	; x = st0

		xor ecx,ecx
		lea edx,points
		ASSUME edx:PTR vertex_s
	C0:	;; store current
		fld st(1)
		fstp [edx+ecx].m_y
		fst [edx+ecx].m_x

		;; compute next	; x			y
		fld st			; x			x		y
		fabs			; |x|		x		y
		fld1			; 1			|x|		x		y
		faddp st(1),st	; 1+|x|		x		y
		fsubrp st(2),st	; x			1-y+|x|
		fxch			; 1-y+|x|	x

		;; adjust bounding box (FCMOVcc)
		fcom st(4)	;min x
		fnstsw ax
		sahf
		ja gtx
		fld st
		fxch st(5)
		fstp st
	gtx:fcom st(2)	;max x
		fnstsw ax
		sahf
		jb ltx
		fld st
		fxch st(3)
		fstp st
	ltx:fld st(1)	;min y
		fcom st(6)
		fnstsw ax
		sahf
		ja gty
		fxch st(6)
	gty:fst st		;max y
		fcom st(4)
		fnstsw ax
		sahf
		jb lty
		fxch st(4)
	lty:fstp st
		add ecx,SIZEOF vertex_s
		cmp ecx,SIZEOF points ; array index
		jne C0

		;; remember current as initial for next draw
		fstp x0
		fstp y0

		;; normalize to [0,1] and set colors
		fsub st,st(2) ;max x - min x
		fld st(3)
		fsubp st(2),st ;max y - min y
		fcom
		fnstsw ax
		sahf
		jb igt
		fxch
	igt:fstp st
		fld1
		fdivrp st(1),st ; scale
		
		xor ecx,ecx
		lea edx,points
		ASSUME edx:PTR vertex_s
	i0:	fld [edx+ecx].m_x
		fsub st,st(2)
		fmul st,st(1)
		;
		push eax
		fld st
		fmul __255
		fistp DWORD PTR [esp]
		pop eax
		;
		fmul __FRAME_WIDTH
		fld [edx+ecx].m_y
		fsub st,st(4)
		fmul st,st(2)
		;
		push eax
		fld st
		fmul __255
		fistp DWORD PTR [esp]
		fild DWORD PTR [esp]
		fmul __65536
		fistp DWORD PTR [esp]
		or [esp],eax
		pop eax
		;
		fld1
		fsubrp st(1),st
		fmul __FRAME_HEIGHT
		fstp [edx+ecx].m_y
		fstp [edx+ecx].m_x
;		mov [edx+ecx].m_z,0.5
		mov [edx+ecx].m_z,01234567h
		org $-4
		REAL4 0.5
;		mov [edx+ecx].m_recip_w,2.0
		mov [edx+ecx].m_recip_w,01234567h
		org $-4
		REAL4 2.0
		mov ah,al
		neg al
		or eax,0ff000000h
		mov [edx+ecx].m_diffuse,eax
		add ecx,SIZEOF vertex_s
		cmp ecx,SIZEOF points
		jne i0
		fstp st ; scale
		fstp st ; min x
		fstp st ; min y

		;; draw the scene and clear the frame buffer
		mcall [g_device],IDirect3DDevice8_BeginScene
		mcall [g_device],IDirect3DDevice8_Clear,0,NULL,D3DCLEAR_TARGET OR D3DCLEAR_ZBUFFER,0,___1f,0

		;; draw them in batches < 64K vertices for really large point counts
		mcall [g_device],IDirect3DDevice8_SetVertexShader,D3DFVF_XYZRHW OR D3DFVF_DIFFUSE
		BATCH_COUNT EQU 0FF00h
		xor ebx,ebx ; current index into array of vertices
	B0:	cmp ebx,NUM_POINTS
		ja B9
	B1:	mov ecx,BATCH_COUNT ; vertices to process
		add ebx,BATCH_COUNT
		cmp ebx,NUM_POINTS
		jbe B2
		sub ecx,ebx
		add ecx,NUM_POINTS
	B2:	sub ebx,BATCH_COUNT
		mov eax,SIZEOF vertex_s
		mul ebx
		lea esi,[points + eax]
		mcall [g_device],IDirect3DDevice8_DrawPrimitiveUP,D3DPT_POINTLIST,ecx,esi,SIZEOF vertex_s
		add ebx,BATCH_COUNT
		jmp B0
	B9:	mcall [g_device],IDirect3DDevice8_EndScene


		;; present the scene & handle lost devices
		mcall [g_device],IDirect3DDevice8_Present,NULL,NULL,NULL,NULL
	L1:	cmp eax,D3DERR_DEVICELOST
		jne L9
		L2:	invoke Sleep,1000 ; damn, I hate waiting
			mcall [g_device],IDirect3DDevice8_TestCooperativeLevel
			cmp eax,D3DERR_DEVICENOTRESET
			jne L2
		mcall [g_device],IDirect3DDevice8_Reset,OFFSET g_presentation
		or eax,0 ;FAILED
		mov eax,D3DERR_DEVICELOST
;		mov eax,D3DERR_DEVICENOTRESET ;
		je L1
	L9:
		;; clear Win32 dirty region
		invoke ValidateRect,g_window,NULL
	@Bye:
		ret
	Frame_Draw ENDP



	WndProc PROC hWnd:DWORD,uMsg:DWORD,wParam:DWORD,lParam:DWORD
	@@:	cmp uMsg,WM_SIZING
		jne @F
		call Frame_Draw
		return 0
	@@: cmp uMsg,WM_PAINT
		jne @F
		call Frame_Draw
		return 0
	@@:	cmp uMsg,WM_DESTROY
		jne @F
		invoke PostQuitMessage,0
		return 0
	@@: cmp uMsg,WM_SYSCOMMAND ; restore doesn't post the above messages
		jne @F
		cmp wParam,SC_RESTORE
		jne @F
		invoke DefWindowProc,hWnd,uMsg,wParam,lParam
		call Frame_Draw
		return 0
	@@:	invoke DefWindowProc,hWnd,uMsg,wParam,lParam
		ret
	WndProc ENDP


	WinMain:
		invoke InitCommonControls	; For WinXP Theme
		invoke GetModuleHandle,NULL	; Get hInstance
		mov hInst,eax

		; Register the window class
		invoke LoadIcon,hInst,IDI_ICON
		mov wc.hIcon,eax
		invoke LoadCursor,NULL,IDC_ARROW
		mov wc.hCursor,eax
		invoke RegisterClassEx,OFFSET wc
 
		; Create the application's window.
		invoke GetDesktopWindow
		invoke CreateWindowEx,NULL,OFFSET APP_CLASS,OFFSET APP_NAME,WS_OVERLAPPEDWINDOW,\
				CW_USEDEFAULT,CW_USEDEFAULT,FRAME_WIDTH,FRAME_HEIGHT,\
				eax,NULL,hInst,NULL
		mov g_window,eax

		;; Get IDirect3D8 interface, then create a device
		invoke Direct3DCreate8, D3D_SDK_VERSION
		or g_d3d,eax
		je Bye

		mcall [g_d3d],IDirect3D8_CreateDevice,\
				D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,g_window,\
				D3DCREATE_SOFTWARE_VERTEXPROCESSING,\
				OFFSET g_presentation, OFFSET g_device
		push eax
		mcall [g_d3d],IDirect3D8_Release
		pop eax
		cmp eax,D3D_OK
		jne Bye
		and g_d3d,0

		; Show the window
		invoke ShowWindow,g_window,SW_SHOWDEFAULT
		invoke UpdateWindow,g_window

	MsgLoop:
		invoke PeekMessage,OFFSET msg,NULL,0,0,PM_REMOVE
		test eax,eax
		je MsgLoop
		cmp msg.message,WM_QUIT
		je MsgExit
		invoke TranslateMessage,OFFSET msg
		invoke DispatchMessage,OFFSET msg
		jmp MsgLoop
	MsgExit:
		invoke PostQuitMessage,msg.wParam
		
		or g_device,0
		je Bye
		mcall [g_device],IDirect3DDevice8_Release

	Bye:
		invoke UnregisterClass,OFFSET APP_CLASS,hInst
		invoke ExitProcess,msg.wParam

_TEXT ENDS

END WinMain

